"""An alternative TerminalReporter providing Rich text based output."""
from __future__ import annotations

import importlib.util
import os
import threading
import time
import weakref
from pathlib import Path
from typing import Callable, Literal, TYPE_CHECKING, cast

import pytest
from _pytest.config import ExitCode
from _pytest.python import Function

from rich.padding import Padding
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table

from . import compat
from .header import generate_header_panel
from .helpers import Helper, NodeResult, ReporterBase
from .progress import (
    DescriptiveProgressBar, GraphBar, PercentageProgressBar,
    SlottedProgressBar, TestProgressMapper)

if TYPE_CHECKING:
    import warnings
    from collections.abc import Iterator, Sequence

    from _pytest.reports import BaseReport, CollectReport, ExceptionRepr
    from _pytest.terminal import TerminalReporter

HAVE_XDIST = importlib.util.find_spec('xdist') is not None
HORIZONTAL_PAD = (0, 1, 0, 1)

summary_exit_codes = {
    ExitCode.OK,
    ExitCode.TESTS_FAILED,
    ExitCode.INTERRUPTED,
    ExitCode.USAGE_ERROR,
    ExitCode.NO_TESTS_COLLECTED,
}

# This is here for ad-hoc debugging. Any strings in this list get printed in
# the final summary box.
messages: list[str] = []


def in_main_thread():
    """Test if this is the main (driver) thread."""
    return threading.main_thread().ident == threading.current_thread().ident


def interpret_report(report: BaseReport) -> tuple[str, str]:
    """Interpret some details of a report.

    This is here to provide compatability with the standard terminal reporter.
    It is largely cut-and-paste from Pytest code.

    :return:
        A tuple of strings; outcome, letter-code.
    """
    letter = 'F'
    if report.passed:
        letter = '.'
    elif report.skipped:
        letter = 's'

    outcome: str = report.outcome
    if report.when in ('collect', 'setup', 'teardown') and outcome == 'failed':
        outcome = 'error'
        letter = 'E'

    return outcome, letter


class TimeStatCollector:
    """A collector of timing statistics."""

    def __init__(self):
        self.stats: dict[str, list[float, float]] = {}

    def start(self, name):
        """Start timing something."""
        self.stats[name] = [time.time(), None]

    def stop(self, name):
        """Stop timing something."""
        if name in self.stats and self.stats[name][1] is None:
            self.stats[name][1] = time.time()

    def __iter__(self) -> Iterator[tuple[str, float]]:
        """Iterate over the timeing stats."""
        for name, (start, stop) in self.stats.items():
            if stop is not None:
                yield name, stop - start


class Collection(Helper):
    """Information about and management of collected tests."""

    # pylint: disable=too-many-instance-attributes
    prog_mapper: TestProgressMapper

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.failed: dict[str, CollectReport] = {}
        self.skipped: dict[str, Collection] = {}
        self.items: dict[str, pytest.Item] = {}
        self.deselected: set[str] = set()
        self.collection_thread: threading.Thread | None = None
        self.processed_reports: set[str] = set()
        self.active_xdist_nodes: int | None = self.numprocesses
        self.started = False

    def report_already_seen(self, report) -> bool:
        """Check if a report has already been seen.

        :return: True if the report has already been seen.
        """
        seen = report.nodeid in self.processed_reports
        self.processed_reports.add(report.nodeid)
        return seen

    def start(self):
        """Prepare for the start of test collection."""
        self.progress.add_bar(
            'collection', bar=DescriptiveProgressBar(label='Collection'))

        if self.numprocesses is not None:
            # Looks like we are running under pytest-xdist. Do an extra
            # parallel collecion so that we have full test details. The
            # pytest-xdist plugin 'optimizes' away collection information so
            # this is the only way to provide real-time collection stats,
            # albeit a somewhat misleading ones.
            #
            # The separate thread allows the pytest-xdist driver to continue
            # managing its child processes.
            def do_collect():
                pts = self.pytest_session
                pts.items = pts.perform_collect()

            self.collection_thread = threading.Thread(
                target=do_collect, name='par-collect')
            self.collection_thread.start()
        self.started = True

    if HAVE_XDIST:
        def xdist_node_finished(self, node, ids: list[str]):
            """Process pytest-xdist node collection complete notification."""
            self.active_xdist_nodes -= 1
            if self.active_xdist_nodes == 0:
                self.cleanup()
                session = self.pytest_session
                if self.failed and not self.continue_on_error:
                    n_failures = len(self.failed)
                    plural = 's' if n_failures > 1 else ''
                    msg = f'{n_failures} error{plural} during collection'
                    raise session.Interrupted(msg)

    def finish(self):
        """Process the completion of test collection.

        When running an extra, parallel connection (because pytest-xdist is on
        control of test collection), this gets invoked in the collection
        thread, so joining of the thread must be performed by other code
        invoking the `cleanup` method.
        """
        if self.progress is not None:
            self.update_progress(final=True)
        self.prog_mapper = TestProgressMapper(self, self.items)

    def cleanup(self) -> None:
        """Clean up after collection.

        Called by various reporting hooks because the `finish` is not in a
        position to join for the collection threead.
        """
        if self.collection_thread:
            self.collection_thread.join()
            self.collection_thread = None

    def handle_report(self, report: pytest.CollectReport) -> None:
        """Handle a report on test collection progress.

        This is normally invoked directly by the PyTest framework, but
        pytest-xdist will also invoke this when collection errors occurr. So we
        need to handle multiple reports with the same ``report.nodeid`` in a
        graceful manner.

        :report:
            The is a pytest.CollectReport object. The important attributes are:

            result
                A list of pytest.Item and pytest.Collector instances. The
                pytest.Item instances represent actual tests.
            outcome
                A string that can be 'passed', 'skipped or 'failed'.
            longrepr
                If the outcome 'skipped' then this is as a tuple of:

                    path_name: str, lineno: int, message: str

                If outcome is 'failed' then this is either an
                ExceptionChainRepr or a simple CollectErrorRepr. The latter
                simply contains a formatted exception in its ``longrepr``
                string attribute.
        """
        item_types = pytest.Item, Function
        if not self.report_already_seen(report):
            if report.outcome == 'failed':
                self.failed[report.nodeid] = report
            elif report.outcome == 'skipped':
                self.skipped[report.nodeid] = report
            else:
                items = [
                    x for x in report.result if isinstance(x, item_types)]
                for item in items:
                    self.items[item.nodeid] = item
                self.update_progress(final=False)

    def deselect(self, items: Sequence[pytest.Item]) -> None:
        """Handle deselection of one or more tests."""
        for item in items:
            self.deselected.add(item.nodeid)
            self.items.pop(item.nodeid)

    def update_progress(self, *, final: bool = False):
        """Update the test collection progress."""
        if self.progress is None:
            return                                          # pragma: defensive

        deselected = len(self.deselected)
        failed = len(self.failed)
        skipped = len(self.skipped)
        selected = len(self.items) - deselected
        leader = 'complete' if final else 'running'
        desc = f'[cyan][bold]{leader}[/]: [/]selected=[green]{selected}[/]'
        if skipped:
            desc += f' skipped=[blue]{skipped}[/]'
        if deselected:
            desc += f' deselected=[yellow]{deselected}[/]'
        if failed:
            desc += f' failed=[red]{failed}[/]'

        self.progress.update('collection', text=desc, refresh=final)

    def report_summary(self):
        """Report a summary of any test collection failures."""
        self.report_failures(
            self.failed.values(), typename='collection',
            title='Collection Errors', style='')

    @property
    def continue_on_error(self):
        """The value of the global continue_on_collection_errors option."""
        return self.pytest_session.config.option.continue_on_collection_errors


class RunPhase(Helper):
    """Information about and management of the test execution phase."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.node_to_result: dict[str, NodeResult] = {}
        self.path_to_items: dict[Path, dict[str, pytest.Item]] = {}
        self.path_width: int = 0
        self.num_tests_executed = 0
        self.running_tests: set[str] = set()
        self.name_width = 10
        self.start_times: dict[str, float] = {}
        self.exec_times: dict[str, float] = {}
        self.svg_created = False

    @property
    def group_by_dir(self):
        """True if tests are group by directory."""
        return self.collection.group_by_dir

    def start(self):
        """Prepare for the test execution phase."""
        if self.collection.failed:
            return

        # Create a mapping from test script path to another dict mapping nodeid
        # to test item.
        for nodeid, item in self.collection.items.items():
            self.node_to_result[nodeid] = NodeResult(weakref.proxy(self))
            path, _ = self.parse_nodeid(nodeid)
            if path not in self.path_to_items:
                self.path_to_items[path] = {}
            self.path_to_items[path][nodeid] = item

        # Add a progress bar for each grouping of tests.
        prog_mapper = self.collection.prog_mapper
        self.name_width = prog_mapper.name_width
        for name, items in prog_mapper.item_map.items():
            label = prog_mapper.label_map.get(name, name)
            bar = SlottedProgressBar(label=label, nitems=len(items))
            self.progress.add_bar(name, bar=bar)

        # Add progress bars for the overall progress and current number of
        # active tes runners, The latter only if pytest-xdist is active.
        bar = PercentageProgressBar(
            label='Overall', nitems=len(self.node_to_result))
        self.progress.add_bar('overall', bar=bar)
        if self.numprocesses is not None and self.numprocesses > 1:
            bar = GraphBar(
                label='Parallel test count', max_count=self.numprocesses)
            self.progress.add_bar('running', bar=bar)

    def log_test_start(
            self, nodeid: str, location: tuple[str, int | None, str],
        ) -> None:
        """Note the starting of a given test."""
        if nodeid in self.deselected:
            return                                          # pragma: defensive
        result: NodeResult
        result, name = self.get_node_result(nodeid)
        if not result:
            return                                          # pragma: defensive

        result.started = True
        self._update_progress(name)
        self.running_tests.add(name)
        self.start_times[nodeid] = time.time()

    def log_report(self, report: pytest.TestReport) -> None:
        """Handle a report for a given running test.

        This is invoked multiple times for a given test The ``when`` attribute
        identifies the test's phase; 'setup', 'call' and 'teardown'. This is
        always invoked for the setup and teardown phases, but is omitted for
        the call phase if the setup phase failed.
        """
        nodeid = report.nodeid
        if nodeid in self.deselected:
            return                                          # pragma: defensive

        result: NodeResult
        result, name = self.get_node_result(nodeid)
        if result:
            setattr(result, report.when, report)
            self._update_progress(name)

    def log_test_end(self, nodeid: str) -> None:
        """Note the completion of one test."""
        if nodeid in self.start_times:
            self.exec_times[nodeid] = time.time() - self.start_times[nodeid]
        self.num_tests_executed += 1
        _result, name = self.get_node_result(nodeid)
        self.running_tests.discard(name)

    def _update_progress(self, nodeid: str = '', *, refresh: bool = False):
        result: NodeResult
        result, name = self.get_node_result(nodeid)
        if not result:
            return                                          # pragma: defensive

        group_name = self.collection.prog_mapper.group(name)
        path_items = self.collection.prog_mapper.items_for_test_group(name)

        results = [self.node_to_result[item.nodeid] for item in path_items]
        tot_finished = len([res for res in results if res.finished])
        idx = results.index(result)
        self.progress.update(
            group_name, count=tot_finished, idx=idx, char_str=result.indicator)

        tot_finished = len([
            res for res in self.node_to_result.values() if res.finished])
        self.progress.update('overall', count=tot_finished)
        self.progress.update(
            'running', count=len(self.running_tests),
            refresh=tot_finished >= len(self.node_to_result))

    def finish(self):
        """Handle completion of the execution phase."""
        if self.progress:
            self.running_tests = {}
            self.progress.update('running', count=0)
            self.progress.stop()
        times_path_name = self.config.getoption('rich_store_exec_times')
        if times_path_name:
            times = [(t, nodeid) for nodeid, t in self.exec_times.items()]
            with Path(times_path_name).open('wt', encoding='utf-8') as f:
                for t, nodeid in sorted(times, reverse=True):
                    f.write(f'{t:.3f} {nodeid}\n')

    def report_summary(self):
        """Report a summary of any run-phase failures."""
        self.report_failures(
            self.failed, 'run-phase', title='Test Failures',
            style=self.style_map.get('error', ''))
        self.report_failures(
            self.setup_errored, 'run-phase', title='Setup Errors',
            style=self.style_map.get('setup_errored', ''))
        self.report_failures(
            self.teardown_errored, 'run-phase', title='Teardown Errors',
            style=self.style_map.get('teardown_errored', ''))

    def get_node_result(self, nodeid: str) -> tuple[NodeResult | None, str]:
        """Get the result for a given node ID.

        When running under pytest-xdist with 'grouped' tests the reported ID
        may have an append '@<group-name>'. This method handles removing the
        group name part as required.
        """
        name = nodeid
        result = self.node_to_result.get(nodeid)
        if not result:
            root, *_ = nodeid.rpartition('@')
            result = self.node_to_result.get(root)
            if result:
                name = root
        return result, name

    @property
    def collection(self) -> bool:
        """The test collection information."""
        return self.reporter.collection

    @property
    def deselected(self) -> set[str]:
        """The nodeid set of deselected tests."""
        return self.reporter.collection.deselected

    @property
    def not_run(self) -> Sequence[NodeResult]:
        """The results for tests that did not run (no teardown)."""
        return [res for res in self.node_to_result.values()
            if not res.teardown]

    @property
    def passed(self) -> Sequence[NodeResult]:
        """The results for the passing tests."""
        return [res for res in self.node_to_result.values()
            if res.passed_report and not res.xpassed_report
                and res.finished]

    @property
    def setup_errored(self) -> Sequence[NodeResult]:
        """The results for the tests failed during setup."""
        return [res for res in self.node_to_result.values()
            if res.setup_error_report]

    @property
    def teardown_errored(self) -> Sequence[NodeResult]:
        """The results for the tests that failed during teardown."""
        return [res for res in self.node_to_result.values()
            if res.teardown_error_report]

    @property
    def failed(self) -> Sequence[NodeResult]:
        """The results for the tests that failed."""
        return [res for res in self.node_to_result.values()
            if res.failed_report and not res.xfailed_report
                and not res.setup_error_report]

    @property
    def skipped(self) -> Sequence[NodeResult]:
        """The results for the tests that were skipped."""
        return [res for res in self.node_to_result.values()
            if res.skipped_report and not res.xfailed_report]

    @property
    def xfailed(self) -> Sequence[NodeResult]:
        """The results for the tests that failed as expected."""
        return [res for res in self.node_to_result.values()
            if res.xfailed_report]

    @property
    def xpassed(self) -> Sequence[NodeResult]:
        """The results for the tests that failed as expected."""
        return [res for res in self.node_to_result.values()
            if res.xpassed_report]


class Session(Helper):
    """Information about and management of a test session."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.interrupt = None
        self.start_time: float = time.time()

    def start(self, session: pytest.Session) -> None:
        """Handle the start of the session.

        This is the first notification; it occurs before test collection.
        """
        if not self.no_header:
            header = generate_header_panel(session)
            self.console.print(header)

    def finish(self, exitstatus: int | pytest.ExitCode) -> None:
        """Handle the completion of a test session."""
        self.collection.cleanup()

        if self.collection.failed:
            self.collection.report_summary()
        if exitstatus == ExitCode.INTERRUPTED:
            self.console.print(f'Run was interrupted: {self.interrupt.value}')

        if not compat.run_phase_suppressed(self.config):
            self.run_phase.report_summary()
            self.print_summary()
            # TODO: Check that we replicate all necessary
            #       pytest_terminal_summary output.
            if exitstatus in summary_exit_codes and not self.no_summary:
                self.config.hook.pytest_terminal_summary(
                    terminalreporter=self.reporter, exitstatus=exitstatus,
                    config=self.config)

    # TODO: Set up color scheme.
    def print_summary(self):
        """Print the general summary of the test run."""
        def add_row(value, text, style):
            a = Padding(str(value), pad=HORIZONTAL_PAD, style=style)
            b = Padding(text, pad=HORIZONTAL_PAD, style='default')
            table.add_row(a, b, style='default')

        table = Table.grid()
        table.add_column(justify='right', no_wrap=True)
        table.add_column(no_wrap=True)
        table.add_column(no_wrap=True)

        add_row(
            len(self.collection.items) + len(self.collection.deselected),
            'Total tests', 'default')
        self._add_test_summaries(add_row)

        for name, t in self.reporter.time_stats:
            add_row(f'{t:6.2f}s', name, 'default')
        elapsed = f'{time.time() - self.start_time:6.2f}s'
        add_row(elapsed, 'Overall', 'default')
        for m in messages:
            add_row('', m, 'default')                        # pragma: no cover

        panel = Panel(
            table, title='Summary', style='bold blue', expand=False,
            border_style='bold blue')
        self.console.print(panel)

    def _add_test_summaries(self, add_row: Callable[[int, str, str], None]):
        """Add the summaries for test results, passed, failed, *etc*."""
        if self.collection.deselected:
            add_row(len(self.collection.deselected), 'Deselected',
            self.style_map.get('deselected', ''))
        type_names = (
            'passed', 'failed', 'xfailed', 'xpassed', 'not_run', 'skipped',
            'setup_errored', 'teardown_errored')
        name_to_label = {
            'xfailed': 'Expected failures',
            'xpassed': 'Unexpected passes',
            'setup_errored': 'Setup errors',
            'teardown_errored': 'Teardown errors',
        }
        for state in type_names:
            if seq := getattr(self.run_phase, state):
                label = name_to_label.get(
                    state, state.capitalize().replace('_', ' '))
                indicator = self.indicator_map.get(state, '?')
                style = self.style_map.get(state, '')
                if indicator:
                    if style:
                        label = f'{label} ([{style}]{indicator}[/])'
                    else:
                        label = f'{label} ({indicator})'
                add_row(len(seq), label, style)

    def handle_keyboard_interrupt(
            self, excinfo: pytest.ExceptionInfo[BaseException]) -> None:
        """Handle a keyboard or rleated interrupt."""
        self.interrupt = excinfo

    @property
    def no_header(self) -> bool:
        """The no-header command line option value."""
        return self.config.getoption('no_header')

    @property
    def collection(self) -> RunPhase:
        """The Collection handler."""
        return self.reporter.collection

    @property
    def run_phase(self) -> RunPhase:
        """The RunPhase handler."""
        return self.reporter.run_phase


class RichTerminalReporter(ReporterBase):
    """A replacement for the standard pytest terminal reporter.

    This needs to provide quite a large number of hook methods. Most of them
    simply hand off to `Helper` based objects.
    """

    # pylint: disable=too-many-public-methods
    Status = Literal['collected', 'running', 'success', 'fail', 'error']

    def __init__(self, config: pytest.Config, std_reporter: TerminalReporter):
        super().__init__(config)
        self.std_reporter = std_reporter
        self.pytest_session = None
        self.collection = Collection(self)
        self.run_phase = RunPhase(self)
        self.session = Session(self)
        self.monkey_patch_terminal_reporter()
        self.time_stats = TimeStatCollector()
        self.time_stats.start('Init phase')

        # This is required to support some standard pytest features. Currently
        # I know that the --duration option makesuse of this.
        self.stats: dict[str, list] = {}

    def monkey_patch_terminal_reporter(self):
        """Patch parts of the terminal reporter.

        As far as I can tell, it is basically impossible to completely override
        that standard pytest terminal reporter in a completely clean way.

        This is my practical solution.
        """
        self.std_reporter.write = self.write
        self.std_reporter.rewrite = self.rewrite
        self.std_reporter.write_line = self.write_line

    @property
    def _tw(self) -> pytest.TerminalWriter:
        """The terminal write emulation.

        This provides compatability for other parts of pytest and plugins.
        """
        return cast(pytest.TerminalWriter, self.formatter)

    ## Overall session management.
    @pytest.hookimpl(trylast=True)
    def pytest_sessionstart(self, session: pytest.Session) -> None:
        """Perform required actions at the start of the session."""
        self.pytest_session = session
        self.session.start(session)
        if os.environ.get('PYTEST_RICHER_FORCE_ERROR', ''):
            msg = 'Just for testing'
            raise RuntimeError(msg)

    @pytest.hookimpl(tryfirst=True)
    def pytest_sessionfinish(
            self, session: pytest.Session, exitstatus: int | pytest.ExitCode,
        ) -> None:
        """Perform required actions at the end of the session."""
        self.time_stats.stop('Execution')
        self.run_phase.finish()
        self.session.finish(exitstatus)

    @property
    def numprocesses(self) -> int | None:
        """The number of pytest-xdist processes.

        :return:
            The value ``None`` indicates that pytest-xdist is not being used.
            Otherwise this is an integer >= 1.
        """
        return self.config.getoption('numprocesses', None)

    if HAVE_XDIST:
        @pytest.hookimpl(trylast=True)
        def pytest_xdist_node_collection_finished(
                self, node: str, ids: list[str]) -> None:
            """Handle a pytest-xdist worker completing its collection phase."""
            self.collection.xdist_node_finished(node, ids)

    @pytest.hookimpl(tryfirst=True)  # after _pytest.runner
    def pytest_collection(self) -> None:
        """Prepare for test collection.

        This is invoked just before collection starts.
        """
        if not self.collection.started:
            self.time_stats.stop('Init phase')
            self.time_stats.start('Collection')
            self.collection.start()

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtestloop(self):
        """Prepare for the start of test execution.

        When pytest-xdist is not active, this is invoked after test collection,
        just before the tests start executing. When pytest-xdist is active this
        is invoked too early to be of use.

        TODO: So this can be deleted.
        """

    def pytest_collection_finish(self, session: pytest.Session) -> None:
        """Handle the completion of test collection.

        When running with pytest-xdist, this is epected to occur in the
        'per-collect' background thread. Otherwise this occurs in the main
        thread.
        """
        self.time_stats.stop('Collection')
        self.time_stats.start('Execution')
        self.collection.finish()
        if not compat.run_phase_suppressed(self.config):
            self.run_phase.start()
        else:
            self.progress.stop()

    def pytest_collectreport(self, report: pytest.CollectReport) -> None:
        """Handle report of zero or more collected tests.

        The report can indicate collection failure.
        """
        self.pytest_collection()
        self.collection.handle_report(report)

    def pytest_internalerror(
            self,
            excrepr: ExceptionRepr,
            excinfo: pytest.ExceptionInfo[BaseException],
        ) -> None | bool:
        """Report an internal error."""
        self.console.print(self.session.format_internal_error(excinfo))
        return True

    def pytest_warning_recorded(
            self,
            warning_message: warnings.WarningMessage,
            nodeid: str) -> None:
        """Note standard Python warning.

        Currently we are just dropping warnings.
        """

    def pytest_deselected(self, items: Sequence[pytest.Item]) -> None:
        """Not the deselction of one or more tests."""
        self.collection.deselect(items)

    def pytest_runtest_logstart(
            self, nodeid: str, location: tuple[str, int | None, str]) -> None:
        """Log the start of a test."""
        self.run_phase.log_test_start(nodeid, location)

    def pytest_runtest_logfinish(
            self, nodeid: str, location: tuple[str, int | None, str]) -> None:
        """Log the completion of a test."""
        if in_main_thread():
            self.run_phase.log_test_end(nodeid)

    def pytest_runtest_logreport(self, report: pytest.TestReport) -> None:
        """Log the reort for a test run."""
        self.run_phase.log_report(report)
        category, _code = interpret_report(report)
        self._save_reports(category, [report])

    def pytest_keyboard_interrupt(
            self, excinfo: pytest.ExceptionInfo[BaseException]) -> None:
        """Handle KeyboardInterrupt base exceptions.

        Note that pytest uses this for non-keyboard interruptions, such as
        errors during test collection.
        """
        self.session.handle_keyboard_interrupt(excinfo)

    def pytest_unconfigure(self) -> None:
        """Perform final action before exiting."""

    def write_sep(
            self, sep: str, title: str | None = None,
            fullwidth: int | None = None, **markup: bool,
        ) -> None:
        """Process request to write a ruled separator line.

        :sep:       Nominally a separator character, but it may be more than
                    one character.
        :title:     A title to embed within the separator line.
        :fullwidth: A value to over-ride the terminal width.
        :markup:    Markup flag values.
        """
        style_names = self.session.reporter.convert_markup(markup)
        style = ('[' + ' '.join(style_names) + ']') if style_names else ''
        self.progress.handle_output(
            self.console.print, Rule(title, style=style, characters=sep))

    def write(self, text: str | bytes, **markup: bool) -> None:
        """Process a request to write some text or bytes."""
        if not isinstance(text, str):
            text = str(text, errors='replace')
        self.progress.handle_output(
            self.console.print, self.formatter.convert_ansi_codes(text),
            end='')

    def write_line(self, line: str | bytes, **markup: bool) -> None:
        """Process a request write a line of text or bytes."""
        if not isinstance(line, str):
            line = str(line, errors='replace')
        self.progress.handle_output(
            self.console.print, self.formatter.convert_ansi_codes(line))

    # TODO: What about when progress is not running?
    # TODO: Should this start an idle progress display?
    def rewrite(self, line: str, **markup: bool) -> None:
        """Process a request to over-write the current line.

        As far as I can tell, this is used for a "poor man's progress display".
        As such, we simply add another progress task on demand.
        """
        self.progress.add_bar(
            '_extra_', bar=DescriptiveProgressBar(label='Information'))
        self.progress.update('_extra_', text=line.rstrip().replace('\n', ''))

    def _save_reports(self, category: str, reports: list[pytest.TestReport]):
        """Save one or more reports under a given category."""
        if category not in self.stats:
            self.stats[category] = []
        self.stats[category].extend(reports)



"""Progress display support."""
from __future__ import annotations

import re
import time
import weakref
from functools import partial
from typing import TYPE_CHECKING

from rich.live import Live
from rich.markup import escape

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Callable

    import pytest
    from rich.console import Console

    from .helpers import Helper


def long_enough(max_time: float) -> Iterator[bool]:
    """Generate True each time 'max_time' has elapsed."""
    prev_trigger = time.time()
    while True:
        now = time.time()
        if now - prev_trigger >= max_time:
            yield True
            prev_trigger = now
        else:
            yield False


class TestProgressMapper:
    """Management mapping from test nodeid to progress display bars."""

    def __init__(self, helper: Helper, items: dict[str, pytest.Item]):
        self.item_to_group_name: dict[str, str] = {}
        self.item_map: dict[str, list[pytest.Item]] = {}
        self.is_cont: set[str] = set()
        self.label_map: dict[str, str] = {}

        # Try grouping tests by test script.
        item_map = self.item_map
        root = helper.config.rootpath
        for item in items.values():
            name = str(item.path.relative_to(root))
            if name not in item_map:
                item_map[name] = []
            item_map[name].append(item)
            self.item_to_group_name[item.nodeid] = name
        self._split_long_groups(helper.console)

        # If there are too many groups for the screen height, try grouping by
        # directory name.
        if len(item_map) > helper.console.size.height - 6:
            self.item_map.clear()
            self.label_map.clear()
            self.is_cont.clear()
            for item in items.values():
                name = str(item.path.relative_to(root).parent)
                if name not in item_map:
                    item_map[name] = []
                item_map[name].append(item)
                self.item_to_group_name[item.nodeid] = name
            self._split_long_groups(helper.console)

        def get_key(name):
            m = re.match(r'(.*)\[(\d+)\]$', name)
            if m:
                return m.group(1), int(m.group(2))
            else:
                return name, 0                               # pragma: no cover

        self.item_map = {n: item_map[n] for n in sorted(item_map, key=get_key)}

    def _split_long_groups(self, console: Console):
        """Split groups that will not fit on a single line."""
        item_map = self.item_map
        space = console.size.width - self.name_width - 9
        for name, items in list(item_map.items()):
            if len(items) > space:
                item_map.pop(name)
                n = 1
                old_items = list(items)
                while old_items:
                    sub_name = f'{name}[{n}]'
                    item_map[sub_name], old_items = (
                        old_items[:space], old_items[space:])
                    for item in item_map[sub_name]:
                        self.item_to_group_name[item.nodeid] = sub_name
                    if n == 1:
                        self.label_map[sub_name] = name
                    else:
                        self.label_map[sub_name] = ''
                    n += 1

    @property
    def name_width(self) -> int:
        """The width required to display group names."""
        return max((len(str(p)) for p in self.item_map), default=10)

    def group(self, nodeid: str) -> str:
        """Find the name of the group for a given test."""
        return self.item_to_group_name[nodeid]

    def items_for_test_group(self, nodeid: str) -> str:
        """Find test items in the group for a given test."""
        return self.item_map[self.group(nodeid)]


class ProgressBar:
    """Base for Paul's new, hopefully faster, progress bars."""

    parent: ProgressDisplay

    def __init__(self, *, label: str):
        self.label = label
        self._dirty = True
        self._cache = ''

    @property
    def label_width(self) -> int:
        """The width required for the label column."""
        return self.parent.label_width

    @property
    def perc_width(self) -> int:
        """The width required for percant complete column."""
        return self.parent.perc_width

    def render(self) -> str:
        """Render this bar as a line of text."""
        if self._dirty:
            perc = self.render_perc()
            label = f'{escape(self.label):<{self.label_width}}'
            self._cache = f'{perc}{label} {self.render_dynamic()}'
        return self._cache

    def render_perc(self) -> str:
        """Render the percent complete section of this bar."""
        return ' ' * self.perc_width

    def render_dynamic(self) -> str:              # pylint: disable=no-self-use
        """Render the dynamic section of this bar."""
        return ''


class PercentageProgressBar(ProgressBar):
    """Paul's new, hopefully faster, percantage progress bar."""

    def __init__(self, *, nitems: int, label: str):
        super().__init__(label=label)
        self.nitems = nitems
        self.count = 0

    def update(self, count: int) -> None:
        """Update this progress bar's count."""
        if self.count != count:
            self.count = count
            self._dirty = True

    def render_perc(self) -> str:
        """Render the percent complete section of this bar."""
        if self.count >= self.nitems:
            return '[cyan][100%][/cyan] '
        else:
            p = int(min(self.count / self.nitems * 100.0, 99.0))
            return f'[cyan][{p:<3}%][/cyan] '


class SlottedProgressBar(PercentageProgressBar):
    """Paul's new, hopefully faster, slotted progress bar."""

    def __init__(self, *, nitems: int, label: str):
        super().__init__(label=label, nitems=nitems)
        self.slots = ['.'] * nitems
        self.count = 0

    def update(self, count: int, *, idx: int, char_str: str) -> None:
        """Update this progress bar's count and progress slot."""
        # pylint: disable=arguments-differ
        super().update(count)
        if self.slots[idx] != char_str:
            self.slots[idx] = char_str
            self._dirty = True

    def render_dynamic(self) -> str:
        """Render the dynamic section of this bar."""
        return ''.join(self.slots)


class DescriptiveProgressBar(ProgressBar):
    """Paul's new, hopefully faster, descriptive progress bar."""

    def __init__(self, *, label: str):
        super().__init__(label=label)
        self.text: str = ''

    def update(self, *, text: str):
        """Update the dynamic text of this progress bar."""
        if self.text != text:
            self.text = text
            self._dirty = True

    def render_dynamic(self) -> str:
        """Render the dynamic section of this bar."""
        return self.text


class GraphBar(ProgressBar):
    """Paul's new, hopefully faster, descriptive graph bar."""

    def __init__(self, *, label: str, max_count: int):
        super().__init__(label=label)
        self.max_count = max_count
        self.count = 0

    def update(self, *, count: int):
        """Update this progress bar's count."""
        if self.count != count:
            self.count = count
            self._dirty = True

    def render_dynamic(self) -> str:
        """Render the dynamic section of this bar."""
        bar_fg = '=' * self.count
        bar = bar_fg.ljust(self.max_count, '-')
        return f'{self.count:<3} |{bar}|{self.max_count}'


class ProgressDisplay:
    """Paul's new, hopefully faster, progress display."""

    def __init__(self, console: Console):
        self.console = console
        self.bars: dict[str, ProgressBar] = {}
        self._label_width = -1
        self._perc_width = -1
        self.time_trigger = long_enough(0.1)
        self.live = Live(
            console=console,
            auto_refresh=False,
            get_renderable=self.get_renderable)
        self.stored_output: list[Callable[[], None]] = []

    def add_bar(self, name: str, *, bar: ProgressBar):
        """Add a bar to this progress display."""
        if name not in self.bars:
            self.bars[name] = bar
            bar.parent = weakref.proxy(self)
            self._label_width = -1
            self._perc_width = -1
            if not self.live.is_started:
                self.start()

    def start(self):
        """Start or resume the live display of this progress bar."""
        if not self.live.is_started:
            self.live.start()

    def stop(self):
        """Stop the live display of this progress bar."""
        if self.live.is_started:
            self.live.stop()
            for func in self.stored_output:
                func()
            self.stored_output[:] = []

    def get_renderable(self) -> str:
        """Get a renderable form of this progress display.

        This is called by the Rich.Live instance.
        """
        lines = [bar.render() for bar in self.bars.values()]
        return '\n'.join(lines)

    def update(self, name: str, *, refresh: bool = False, **kwargs) -> None:
        """Update a given progress bar.

        :name:    The name identifying the ProgressBar.
        :refresh: If set then the display will unconditionally refreshed.
        :kwargs:  The keyword argument required by the progress bar's update
                  method.
        """
        bar = self.bars.get(name)
        if bar:
            bar.update(**kwargs)
        triggered = next(self.time_trigger)
        if triggered or refresh:
            self.live.refresh()

    def handle_output(self, func, *args, **kwargs):
        """Handle a non-progress output request.

        If the live display is active then the request is buffered until live
        output ends.

        :func:   The function that will produuce the output.
        :args:   Positional arguments for the function.
        :kwargs: The keyword arguments for the function.
        """
        if self.live.is_started:
            self.stored_output.append(partial(func, *args, **kwargs))
        else:
            func(*args, **kwargs)

    @property
    def label_width(self):
        """The width required for the label column."""
        if self._label_width < 0:
            self._label_width = max(
                (len(bar.label) for bar in self.bars.values()), default=10)
        return self._label_width

    @property
    def perc_width(self):
        """The width required for percant complete column."""
        if self._perc_width < 0:
            for bar in self.bars.values():
                if isinstance(bar, PercentageProgressBar):
                    self._perc_width = 7
                    break
            else:
                self._perc_width = 0
        return self._perc_width



"""Base helper support."""
from __future__ import annotations

import pickle
import weakref
from contextlib import suppress
from dataclasses import dataclass
from itertools import dropwhile, filterfalse, takewhile
from pathlib import Path
from typing import ClassVar, Literal, TYPE_CHECKING

import _pytest
import pluggy
import pytest
from _pytest.reports import BaseReport, CollectReport
from rich import box
from rich.console import Console, Group
from rich.padding import Padding
from rich.panel import Panel
from rich.traceback import Stack, Traceback

from . import reporting
from .progress import ProgressDisplay

if TYPE_CHECKING:
    from collections.abc import Generator

    from _pytest._code import ExceptionInfo
    from rich.traceback import Trace

FailureReportTypeName = Literal['run-phase', 'collection']

# Directory paths used to remove 'noisy' parts of test failure tracebacks.
# This is largely copied from the pyetst codebase.
pluggy_path = Path(pluggy.__file__.rstrip('oc'))
if pluggy_path.name == '__init__.py':
    pluggy_path = pluggy_path.parent
pytest_paths = [Path(pytest.__file__).parent, Path(_pytest.__file__).parent]

# This style set defined canonocal state names for tests during or after
# completion of execution.
# pylint: disable=use-dict-literal
styles = dict(
    deselected='yellow',
    failed='red',
    not_run='',
    not_started='',
    passed='green',
    running='bold breight_green',
    setup_running='',
    setup_errored='red',
    skipped='',
    teardown_running='',
    teardown_errored='yellow',
    xfailed='magenta',
    xpassed='green4',
)
std_styles = styles.copy()
std_styles.update(dict(
    skipped='yellow',
    teardown_errored='red',
    xfailed='yellow',
    xpassed='yellow',
))
state_to_indicator = dict(
    failed='✕',
    not_run='.',
    not_started='.',
    passed='✔',
    running='r',
    setup_running='↑',
    setup_errored='u',
    skipped='s',
    teardown_running='↓',
    teardown_errored='d',
    xfailed='f',
    xpassed='p',
)
std_state_to_indicator = state_to_indicator.copy()
std_state_to_indicator.update(dict(
    failed='F',
    passed='.',
    setup_errored='E',
    teardown_errored='E',
    xfailed='x',
    xpassed='X',
))
# pylint: enable=use-dict-literal


class ReporterBase:                    # pylint: disable=too-few-public-methods
    """A replacement for the standard pytest terminal reporter.

    This needs to provide quite a large number of hook methods. Most of them
    simply hand off to `Helper` based objects.
    """

    def __init__(self, config: pytest.Config):
        self.config = config
        self.console: Console = Console(highlighter=None)
        self.formatter = reporting.TerminalFormatter(
            self.console,
            highlight_code=self.config.option.code_highlight == 'yes')
        self.progress = ProgressDisplay(self.console)


@dataclass
class NodeResult:
    """Captured information about a test node's results."""

    parent: Helper
    setup: pytest.TestReport | None = None
    call: pytest.TestReport | None = None
    teardown: pytest.TestReport | None = None
    started: bool = False
    expr_to_state: ClassVar[dict[str, str]] = {
        'not self.started' : 'not_started',
        'not self.setup' : 'setup_running',
        'not self.call and not self.teardown' : 'running',
        'not self.teardown' : 'teardown_running',
        'self.xfailed_report' : 'xfailed',
        'self.xpassed_report' : 'xpassed',
        'self.setup_error_report' : 'setup_errored',
        'self.teardown_error_report' : 'teardown_errored',
        'self.failed_report' : 'failed',
        'self.passed_report' : 'passed',
        'self.skipped_report' : 'skipped',
    }

    @property
    def finished(self) -> bool:
        """True when this test node has completed its run."""
        return self.teardown is not None

    @property
    def indicator(self) -> str:
        """A single (styled) character indicator of this test's state."""
        for expr, state in self.expr_to_state.items():
            # pylint: disable=eval-used
            if eval(expr):                                  # noqa: PGH001,S307
                ind = self.parent.indicator_map.get(state, '?')
                style = self.parent.style_map.get(state, '')
                if style:
                    return f'[{style}]{ind}[/]'
                else:
                    return ind
        return '?'                                           # pragma: no cover

    @staticmethod
    def _match_report(
            report: pytest.TestReport | None, outcomes: list[str],
        ) -> pytest.Report | None:
        if report and report.outcome in outcomes:
            return report
        else:
            return None

    @property
    def passed_report(self) -> pytest.TestReport | None:
        """Any passing report for this test."""
        return self._match_report(self.call, ['passed'])

    @property
    def main_error_report(self) -> pytest.TestReport | None:
        """The most significant error report for this test, if any."""
        report = self._match_report(self.call, ['failed'])
        if not report:
            report = self.setup_error_report
        if not report:
            report = self.teardown_error_report
        return report

    @property
    def failed_report(self) -> pytest.TestReport | None:
        """Any failing report for this test."""
        report = self._match_report(self.call, ['failed'])
        if not report:
            report = self.setup_error_report
        return report

    @property
    def setup_error_report(self) -> pytest.TestReport | None:
        """Any setup error report for this test."""
        return self._match_report(self.setup, ['failed'])

    @property
    def teardown_error_report(self) -> pytest.TestReport | None:
        """Any teardown error report for this test."""
        return self._match_report(self.teardown, ['failed'])

    @property
    def xfailed_report(self) -> pytest.TestReport | None:
        """Any test expected failure report."""
        call = self._match_report(self.call, ['failed', 'skipped'])
        if call and getattr(call, 'wasxfail', None) is not None:
            return call
        else:
            return None

    @property
    def xpassed_report(self) -> pytest.TestReport | None:
        """Any test unexpected passed report."""
        call = self._match_report(self.call, ['passed'])
        if call and getattr(call, 'wasxfail', None) is not None:
            return call
        else:
            return None

    @property
    def skipped_report(self) -> pytest.TestReport | None:
        """Any test skipped report."""
        setup = self._match_report(self.setup, ['skipped'])
        return setup or self._match_report(self.call, ['skipped'])


class Helper:
    """Base class for for terminal reporter helpers."""

    def __init__(self, reporter: ReporterBase):
        self.reporter = weakref.proxy(reporter)

    @property
    def config(self):
        """The PyTest configuration object."""
        return self.reporter.config

    @property
    def formatter(self):
        """The TerminalFormatter object."""
        return self.reporter.formatter

    @property
    def console(self):
        """The reporting console."""
        return self.reporter.console

    @property
    def pytest_session(self) -> pytest.Session:
        """The pytest.Session instance."""
        return self.reporter.pytest_session

    @property
    def numprocesses(self) -> int | None:
        """The number of pytest-xdist processes.

        :return:
            The value ``None`` indicates that pytest-xdist is not being used.
        """
        return self.reporter.numprocesses

    @property
    def no_summary(self) -> bool:
        """The pytest configuration no_summary flag."""
        return self.config.getoption('no_summary')

    @property
    def progress(self) -> ProgressDisplay:
        """The ProgressDisplay."""
        return self.reporter.progress

    @property
    def style_map(self) -> dict[str, str]:
        """The state to style lookup table."""
        if self.config.option.rich_std_symbols:
            return std_styles
        else:
            return styles

    @property
    def indicator_map(self) -> dict[str, str]:
        """The state to indicator character map."""
        if self.config.option.rich_std_symbols:
            return std_state_to_indicator
        else:
            return state_to_indicator

    def parse_nodeid(self, nodeid: str) -> tuple[Path, str]:
        """Parse a nodeid into its path and test_id parts."""
        path_name, _, test_id = nodeid.partition('::')
        path = Path(path_name)
        with suppress(ValueError):
            path = path.relative_to(self.config.rootpath)
        return path, test_id

    def report_failure(
            self, report: pytest.TestReport | pytest.CollectReport,
            typename: str):
        """Report failure."""

    def format_failure_report(
            self, report: pytest.TestReport | pytest.CollectReport,
            typename: str) -> Group:
        """Fromat a failure report.

        :report:   A pytest report.
        :typename: Either 'collection' or 'run-phase'.
        """
        s = []
        rich_info = getattr(report, 'rich_info', None)
        if isinstance(report, (pytest.TestReport, pytest.CollectReport)):
            if rich_info:
                trace: Trace = pickle.loads(rich_info)             # noqa: S301
                if isinstance(report, pytest.CollectReport):
                    # Only the last stack is not useful so weo remove the
                    # others. The second stack is be emptied of frames because
                    # they are all pytest or Python import machinery.
                    trace.stacks[:] = trace.stacks[-1:]
                    trace.stacks[0].frames[:] = []
                prune_stacks(trace, self.config)
                s.extend(self._format_rich_trace(trace))
        else:
            # pragma: no cover
            s.append(f'format_failure: unhandled report type {type(report)}')
        return Group(*s)

    def _format_rich_trace(self, trace: Trace) -> Traceback:
        """Format a list of Rich traces."""
        kw = {}
        level = self.config.getoption('std_tbstyle', 'long')
        if level == 'short':
            kw['extra_lines'] = 0
        if self.config.getoption('showlocals', default=False):
            kw['show_locals'] = True
        return [Traceback(trace=trace, **kw)]

    def format_sections(self, report: pytest.TestReport) -> Generator[Panel]:
        """Yield report sub-sections."""
        show_capture = self.config.option.showcapture
        for title, text in report.sections:
            if show_capture == 'no' or (
                    show_capture != 'all' and show_capture not in title):
                continue
            if isinstance(text, str):
                fixed_text: Group = self.formatter.convert_ansi_codes(text)
            else:
                fixed_text = text                           # pragma: defensive
            p = Panel(fixed_text, title=title)
            yield p

    def format_error(
            self, report: BaseReport, typename: FailureReportTypeName,
        ) -> Group:
        """Format a single pytest failure report."""
        s = []
        s.append(f'[cornflower_blue]{report.nodeid}:')
        s.append(Padding(
            self.format_failure_report(report, typename=typename),
            (0, 0, 0, 1)))
        panels = self.format_sections(report)
        paddded_panels = (Padding(panel, (0, 0, 0, 1)) for panel in panels)
        s.extend(paddded_panels)
        return Group(*s)

    def report_failures(
            self,
            failed: list[NodeResult | CollectReport],
            typename: FailureReportTypeName,
            title: str,
            style: str,
        ):
        """Report one or more failures."""
        def get_pytest_report(elem: NodeResult | CollectReport):
            if isinstance(elem, CollectReport):
                return elem
            else:
                return elem.main_error_report

        # Unlike standard pytest reporting, we always show collection when
        # --no-summary is set.
        if self.no_summary:
            failed = [r for r in failed if isinstance(r, CollectReport)]

        if failed:
            group = Group(
                *(self.format_error(get_pytest_report(report), typename)
                    for report in failed))
            panel = Panel(group, title=title, border_style=style)
            self.console.print(panel)

    @staticmethod
    def format_internal_error(excinfo: ExceptionInfo) -> Panel:
        """Format details of an internal error."""
        trace = Traceback.extract(
            exc_type=excinfo.type, exc_value=excinfo.value,
            traceback=excinfo.tb, show_locals=True)
        traceback = Traceback(trace=trace)
        return Panel(
            traceback, title='INTERNAL ERROR', subtitle='INTERNAL ERROR',
            border_style='bold red', box=box.DOUBLE_EDGE)


def prune_stacks(trace: Trace, config: pytest.Config) -> None:
    """Remove unwanted stacks entries from all stacjs in a Trace."""
    for stack in trace.stacks:
        prune_stack(stack, config)


def prune_stack(stack: Stack, config: pytest.Config) -> None:
    """Remove unwanted entries within a Stack.

    * Any pluggy frames are removed unconditionally.
    * Leading and trailing frames that appear to be any of the follwong are
      removed.

      a. No associated source file (a name like '<string>').
      b. A pytest frame.
    """
    def parents(f):
        if not no_file(f):
            return Path(f.filename).parents
        else:
            return []

    # pylint: disable=unnecessary-lambda-assignment
    is_pluggy = lambda f: pluggy_path in parents(f)                # noqa: E731
    no_file = lambda f: '<' in f.filename and '>' in f.filename    # noqa: E731
    is_pytest = lambda f: any(                                     # noqa: E731
        p for p in pytest_paths if p in parents(f))
    unwanted = lambda f: no_file(f) or is_pytest(f)                # noqa: E731
    wanted = lambda f: not unwanted(f)                             # noqa: E731

    frames = filterfalse(is_pluggy, stack.frames)
    frames = dropwhile(unwanted, frames)
    frames = takewhile(wanted, frames)
    stack.frames[:] = list(frames)

    show_locals = config.getoption('showlocals', default=False)
    for frame in stack.frames:
        if not show_locals:
            frame.locals = None



"""A Rich based plugin for pytest."""
from __future__ import annotations

import os
import pickle
import sys
from typing import TYPE_CHECKING

import pytest
from _pytest.runner import (
    CallInfo, pytest_runtest_makereport as make_run_report)
from rich.traceback import Traceback

from .terminal import RichTerminalReporter

if TYPE_CHECKING:
    from _pytest.terminal import TerminalReporter


def pytest_addoption(parser):
    """Add command line option for this plug-in."""
    group = parser.getgroup(
        'rich', 'pytest-richer', after='terminal reporting')
    group.addoption(
        '--rich',
        action='store_true',
        help='Enable rich terminal reporting using pytest-richer.',
    )
    group.addoption(
        '--rich-std-symbols',
        action='store_true',
        help='Use standard symbols for test results progress display.',
    )
    group.addoption(
        '--rich-store-exec-times',
        action='store', metavar='PATH',
        help='Store test execution times in PATH.',
    )


def add_rich_info_to_report(
        report: pytest.Collector | pytest.TestReport,
        call: CallInfo | None = None) -> None:
    """Add a rich_info attribute to a report."""
    report.rich_info = None
    call = call or getattr(report, 'call', None)
    if call and call.excinfo:
        e = call.excinfo
        tb = Traceback.extract(
            exc_type=e.type, exc_value=e.value, traceback=e.tb,
            show_locals=True)

        # We can pickle a Trace, but apparently the xdist serialiser cannot
        # serialise it. So store in pickled form.
        s = pickle.dumps(tb)
        report.rich_info = s


@pytest.hookimpl(hookwrapper=True)
def pytest_make_collect_report(
        collector: pytest.Collector) -> None:                    # noqa: ARG001
    """Create a pytest.CollectReport with added 'rich_info' attribute.

    This replicates the standard pytest code to create the report and simply
    adds the extra attribute. Adding an attribute in this way is OK here
    because the test object is expected to carry arbitrary additional
    attributes.
    """
    # Create the report exactly like the pytest code does.
    result = yield
    add_rich_info_to_report(result.get_result())


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_makereport(
        item: pytest.Item, call: CallInfo[None]) -> pytest.TestReport:
    """Create a pytest.TestReport with added 'rich_info' attribute.

    This replicates the standard pytest code to create the report and simply
    adds the extra attribute. Adding an attribute in this way is OK here
    because the test object is expected to carry arbitrary additional
    attributes.
    """
    # Create the report exactly like the pytest code does.
    report = make_run_report(item, call)
    add_rich_info_to_report(report, call)
    return report


@pytest.hookimpl(tryfirst=True)
def pytest_cmdline_main(config):
    """Over-ride the --tb option value, as far as pytest code is concerned.

    Note the user's chosen tracback style then override it to be 'short'.
    This ensures that pytest code saves the most useful traceback information.
    This plug-in then takes over the job of interpreting the '--tb' choice
    provided on the command line.

    Doing this here makes sure the value is over-ridden before the xdist
    plugin makes any copies.
    """
    config.option.std_tbstyle = config.option.tbstyle
    if config.option.rich:
        config.option.tbstyle = 'short'


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    """Perform one-time configuration for this plug-in.

    This installs a RichTerminalReporter instance if the user has used the
    `--rich`` command line option.
    """
    # If pytest-xdist is active then our reporter must only be installed
    # for the main process; indicated by PYTEST_XDIST_WORKER not being set.
    if os.environ.get('PYTEST_XDIST_WORKER') is None:            # noqa: SIM102
        if sys.stdout.isatty() and config.getoption('rich'):
            manager = config.pluginmanager
            standard_reporter: TerminalReporter = manager.getplugin(
                'terminalreporter')
            reporter = RichTerminalReporter(config, standard_reporter)
            config.pluginmanager.register(reporter, 'richer-terminal-reporter')


@pytest.hookimpl(tryfirst=True)
def pytest_sessionstart(session: pytest.Session) -> None:
    """Disable the standard report just before the test session runs."""
    if os.environ.get('PYTEST_XDIST_WORKER') is None:
        config = session.config
        if sys.stdout.isatty() and config.getoption('rich'):
            standard_reporter = config.pluginmanager.getplugin(
                'terminalreporter')
            if standard_reporter:
                # Monkey patch the standard reporter class to make output
                # producing hook-methods no-operations.
                cls = standard_reporter.__class__
                cls.pytest_collection_finish = nop
                cls.pytest_collection = nop
                cls.pytest_collectreport = nop
                cls.pytest_deselected = nop
                cls.pytest_internalerror = nop
                cls.pytest_runtest_logfinish = nop
                cls.pytest_runtest_logreport = nop
                cls.pytest_runtest_logstart = nop
                cls.pytest_runtest_logstart = nop
                cls.pytest_runtest_logstart = nop
                cls.pytest_runtestloop = nop
                cls.pytest_sessionfinish = nop
                cls.pytest_sessionstart = nop
                cls.pytest_terminal_summary = nop

                # Deregister then register the standard reported plugin so that
                # our monkey patched methods get used as the reporter's hook
                # methods. Note that other code assumes the existence of the
                # standard reporter, so simply deregistering it is not
                # sensible.
                config.pluginmanager.unregister(standard_reporter)
                config.pluginmanager.register(
                    standard_reporter, 'terminalreporter')


def nop(*_args, **_kwargs):
    """Do nothing.

    This is no-operation used to monkey-patch standard reporter methods.
    """



"""Support for handling pytest reports and related types."""
from __future__ import annotations

from contextlib import contextmanager
from typing import ClassVar, TYPE_CHECKING

from rich.ansi import AnsiDecoder
from rich.console import Group
from rich.syntax import Syntax

if TYPE_CHECKING:
    from collections.abc import Sequence
    from rich.console import Console


class TerminalFormatter:
    """An object formatter replacing the pytest.TerminalWriter.

    Many objects involved in pytest reports and reported output are designed to
    format their own output using a TerminalWriter, which is typically provided
    by the standar pytest.TerminalReporter.

    This provides a partial emulation of a TerminalWriter's, allowing the
    toterminal method of objects to be used to format output to that is is
    usable with a Rich.console.
    """

    ansi_decoder = AnsiDecoder()
    markup_map: ClassVar[dict[str, str]] = {
        'black': '[black]',
        'red': '[red]',
        'green': '[green]',
        'yellow': '[yellow]',
        'blue': '[blue]',
        'purple': '[purple]',
        'cyan': '[cyan]',
        'white': '[white]',
        'Black': '[bright_black]',
        'Red': '[bright_red]',
        'Green': '[bright_green]',
        'Yellow': '[bright_yellow]',
        'Blue': '[bright_blue]',
        'Purple': '[bright_purple]',
        'Cyan': '[bright_cyan]',
        'White': '[bright_white]',
        'bold': '[bold]',
        'light': '[dim]',
        'blink': '',                         # No. Just no!
        'invert': '[reverse]',
    }

    def __init__(self, console: Console, *, highlight_code: bool):
        self.console = console
        self.highlight_code = highlight_code
        self.buffer: list[str | Syntax] or None = None

    # Methods provides to suppoer pytest_richer.
    @contextmanager
    def capture(self) -> list[str | Syntax]:
        """Provide a context where output is captured rather than written.

        While the context is active to get the `buffer` attribute contains a
        list of renderables.
        """
        self.buffer = []
        yield self.buffer
        self.buffer = None

    def convert_markup(self, markup: dict[str, bool]) -> str:
        """Convert pytest markup flags to Rich style names."""
        return ''.join(
            self.markup_map.get(name, '')
            for name, flag in markup.items() if flag)

    def convert_ansi_codes(self, text: str) -> Group:
        """Convert ASNSI codes to Rich markup."""
        return Group(*self.ansi_decoder.decode(text))

    #
    # Methods use by pytest ``toterminal`` implementations.
    #
    @property
    def fullwidth(self) -> int:
        """The width of the console."""
        return self.console.size.width

    @property
    def width_of_current_line(self) -> int:
        """Return an estimate of the width so far in the current line."""
        msg = 'Cannot calculate space left on the current line.'
        raise RuntimeError(msg)

    def markup(self, text: str, **markup: bool) -> str:
        """Format text with defined markup."""
        start = self.convert_markup(markup)
        return f'{start}{text}[/]' if start else text

    def sep(
            self, sepchar: str, title: str = '', fullwidth: int = 0,
            **markup: bool,
        ) -> None:
        """Draw a separator line accross the width of the terminal.

        :sepchar:   Nominally a separator character, but it may be more than
                    one character.
        :title:     A title to embed within the separator line.
        :fullwidth: A value to over-ride the terminal width.
        :markup:    Markup flag values.
        """
        fullwidth = self.fullwidth if fullwidth <= 0 else fullwidth
        if title:
            # We are assuming sepchar is a single character here.
            text = f' {title} '.center(fullwidth, sepchar)
        else:
            text = (sepchar * fullwidth)[:fullwidth]
        self.line(text, **markup)

    def write(self, msg: str, *, flush: bool = False, **markup: bool) -> None:
        """Write a string with defined markup."""
        if msg:
            start = self.convert_markup(markup)
            text = f'{start}{msg}[/]' if start else msg
            if self.buffer is not None:
                self.buffer.append(text)
            else:
                self.console.print(text, end='')

    def line(self, s: str = '', **markup: bool) -> None:
        """Write text plus newline, with defined markup."""
        self.write(s, **markup)
        self.write('\n')

    def flush(self) -> None:
        """Flush pending output - a nul opearation for this class."""



"""Code that creates the opening header panel."""

import sys
from typing import Iterable
from typing import Union

import pytest
from _pytest.main import Session
from _pytest.terminal import _plugin_nameversions
from rich.columns import Columns
from rich.console import Group
from rich.panel import Panel


def generate_header_panel(session: Session) -> Panel:
    columns = [
        _generate_sysinfo_col(),
        _generate_root_col(session),
        _generate_plugins_col(session),
        *_generate_header_hook_cols(session),
    ]

    return Panel(Group(*columns))


def _generate_sysinfo_col() -> Columns:
    column = Columns(
        [
            f"platform [green]{sys.platform}",
            f"pytest [cyan]{pytest.__version__}",
            f"python [cyan]{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        ]
    )

    pypy_version_info = getattr(sys, "pypy_version_info", None)
    if pypy_version_info is not None:
        column.add_renderable(f"pypy [cyan]{'.'.join(map(str, pypy_version_info[:3]))}")

    return column


def _generate_root_col(session: Session) -> Columns:
    return Columns([f"root [cyan][bold]{session.config.rootpath}"])


def _generate_plugins_col(session: Session) -> Union[Columns, None]:
    plugins = session.config.pluginmanager.list_plugin_distinfo()

    if plugins is None:
        return None

    return Columns(
        [
            f"plugins [cyan]{', '.join(_plugin_nameversions(plugins))}",
        ]
    )


def _generate_header_hook_cols(session: Session) -> Iterable[Columns]:
    lines = session.config.hook.pytest_report_header(
        config=session.config, start_path=session.config.invocation_params.dir
    )

    for line_or_lines in reversed(lines):
        if isinstance(line_or_lines, str):
            yield Columns([line_or_lines])
        else:
            for line in line_or_lines:
                yield Columns([line])



"""Support for compatability with cor pytest and plugins."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pytest


def run_phase_suppressed(config: pytest.Config) -> bool:
    """Determine if the run phase has been suppressed.

    This tries to work out, based on things liek command line options, whether
    the normal test execution phase will be performed.
    """
    for name in ('showfixtures', 'show_fixtures_per_test'):
        if config.getoption(name, None):
            return True
    return False
